import autolens as al
import math
from skimage import measure
import numpy as np
import pytest

import autofit as af

from test.unit.mock.model import mock_cosmology

grid = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [2.0, 4.0]])


@pytest.fixture(autouse=True)
def reset_config():
    """
    Use configuration from the default path. You may want to change this to set a specific path.
    """
    af.conf.instance = af.conf.default


class TestEinsteinRadiusMass(object):
    def test__radius_of_critical_curve_and_einstein_radius__radius_unit_conversions(
        self
    ):

        cosmology = mock_cosmology.MockCosmology(kpc_per_arcsec=2.0)

        sis_arcsec = al.mass_profiles.SphericalIsothermal(
            centre=(al.Length(0.0, "arcsec"), al.Length(0.0, "arcsec")),
            einstein_radius=al.Length(2.0, "arcsec"),
        )

        radius = sis_arcsec.average_convergence_of_1_radius_in_units(
            unit_length="arcsec", cosmology=cosmology
        )
        assert radius == pytest.approx(2.0, 1e-4)

        radius = sis_arcsec.einstein_radius_in_units(unit_length="arcsec")
        assert radius == pytest.approx(2.0, 1e-4)

        radius = sis_arcsec.einstein_radius_in_units(
            unit_length="kpc", redshift_profile=0.5, cosmology=cosmology
        )
        assert radius == pytest.approx(4.0, 1e-4)

        sis_kpc = al.mass_profiles.SphericalIsothermal(
            centre=(al.Length(0.0, "kpc"), al.Length(0.0, "kpc")),
            einstein_radius=al.Length(2.0, "kpc"),
        )

        radius = sis_kpc.average_convergence_of_1_radius_in_units(unit_length="kpc")
        assert radius == pytest.approx(2.0, 1e-4)

        radius = sis_kpc.einstein_radius_in_units(unit_length="kpc")
        assert radius == pytest.approx(2.0, 1e-4)

        radius = sis_kpc.einstein_radius_in_units(
            unit_length="arcsec", redshift_profile=0.5, cosmology=cosmology
        )
        assert radius == pytest.approx(1.0, 1e-4)

        nfw_arcsec = al.mass_profiles.SphericalNFW(
            centre=(al.Length(0.0, "arcsec"), al.Length(0.0, "arcsec")),
            kappa_s=0.5,
            scale_radius=al.Length(5.0, "arcsec"),
        )

        radius = nfw_arcsec.average_convergence_of_1_radius_in_units(
            unit_length="arcsec"
        )
        assert radius == pytest.approx(2.76386, 1e-4)

        radius = nfw_arcsec.einstein_radius_in_units(
            unit_length="arcsec", cosmology=cosmology
        )
        assert radius == pytest.approx(2.76386, 1e-4)

        radius = nfw_arcsec.einstein_radius_in_units(
            unit_length="kpc", redshift_profile=0.5, cosmology=cosmology
        )
        assert radius == pytest.approx(2.0 * 2.76386, 1e-4)

        nfw_kpc = al.mass_profiles.SphericalNFW(
            centre=(al.Length(0.0, "kpc"), al.Length(0.0, "kpc")),
            kappa_s=0.5,
            scale_radius=al.Length(5.0, "kpc"),
        )

        radius = nfw_kpc.average_convergence_of_1_radius_in_units(
            unit_length="kpc", redshift_profile=0.5, cosmology=cosmology
        )
        assert radius == pytest.approx(2.76386, 1e-4)

        radius = nfw_kpc.einstein_radius_in_units(unit_length="kpc")
        assert radius == pytest.approx(2.76386, 1e-4)

        radius = nfw_kpc.einstein_radius_in_units(
            unit_length="arcsec", redshift_profile=0.5, cosmology=cosmology
        )
        assert radius == pytest.approx(0.5 * 2.76386, 1e-4)

    def test__einstein_mass__radius_unit_conversions(self):

        cosmology = mock_cosmology.MockCosmology(
            kpc_per_arcsec=2.0, critical_surface_density=2.0
        )

        sis_arcsec = al.mass_profiles.SphericalIsothermal(
            centre=(al.Length(0.0, "arcsec"), al.Length(0.0, "arcsec")),
            einstein_radius=al.Length(1.0, "arcsec"),
        )

        mass = sis_arcsec.einstein_mass_in_units(
            unit_mass="angular", cosmology=cosmology
        )
        assert mass == pytest.approx(np.pi, 1e-4)

        mass = sis_arcsec.einstein_mass_in_units(
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert mass == pytest.approx(2.0 * np.pi, 1e-4)

        sis_kpc = al.mass_profiles.SphericalIsothermal(
            centre=(al.Length(0.0, "kpc"), al.Length(0.0, "kpc")),
            einstein_radius=al.Length(2.0, "kpc"),
        )

        mass = sis_kpc.einstein_mass_in_units(
            unit_mass="angular", redshift_profile=0.5, cosmology=cosmology
        )
        assert mass == pytest.approx(4.0 * np.pi, 1e-4)

        mass = sis_kpc.einstein_mass_in_units(
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert mass == pytest.approx(2.0 * np.pi, 1e-4)

        nfw_arcsec = al.mass_profiles.SphericalNFW(
            centre=(al.Length(0.0, "arcsec"), al.Length(0.0, "arcsec")),
            kappa_s=0.5,
            scale_radius=al.Length(5.0, "arcsec"),
        )

        mass = nfw_arcsec.einstein_mass_in_units(
            unit_mass="angular", cosmology=cosmology
        )
        assert mass == pytest.approx(np.pi * 2.76386 ** 2.0, 1e-4)

        mass = nfw_arcsec.einstein_mass_in_units(
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert mass == pytest.approx(2.0 * np.pi * 2.76386 ** 2.0, 1e-4)

        nfw_kpc = al.mass_profiles.SphericalNFW(
            centre=(al.Length(0.0, "kpc"), al.Length(0.0, "kpc")),
            kappa_s=0.5,
            scale_radius=al.Length(5.0, "kpc"),
        )

        mass = nfw_kpc.einstein_mass_in_units(
            unit_mass="angular", redshift_profile=0.5, cosmology=cosmology
        )
        assert mass == pytest.approx(np.pi * 2.76386 ** 2.0, 1e-4)

        mass = nfw_kpc.einstein_mass_in_units(
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert mass == pytest.approx(0.5 * np.pi * 2.76386 ** 2.0, 1e-4)


def mass_within_radius_of_profile_from_grid_calculation(radius, profile):

    mass_total = 0.0

    xs = np.linspace(-radius * 1.5, radius * 1.5, 40)
    ys = np.linspace(-radius * 1.5, radius * 1.5, 40)

    edge = xs[1] - xs[0]
    area = edge ** 2

    for x in xs:
        for y in ys:

            eta = profile.grid_to_elliptical_radii(np.array([[x, y]]))

            if eta < radius:
                mass_total += profile.convergence_func(eta) * area

    return mass_total


class TestMassWithinCircle(object):
    def test__mass_in_angular_units__singular_isothermal_sphere__compare_to_analytic(
        self
    ):

        sis = al.mass_profiles.SphericalIsothermal(einstein_radius=2.0)

        radius = al.Length(2.0, "arcsec")

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)

        sis = al.mass_profiles.SphericalIsothermal(einstein_radius=4.0)

        radius = al.Length(4.0, "arcsec")

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)

    def test__mass_in_angular_units__singular_isothermal__compare_to_grid(self):

        sis = al.mass_profiles.SphericalIsothermal(einstein_radius=2.0)

        radius = al.Length(1.0, "arcsec")

        mass_grid = mass_within_radius_of_profile_from_grid_calculation(
            radius=radius, profile=sis
        )

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )

        assert mass_grid == pytest.approx(mass, 0.02)

    def test__radius_units_conversions__mass_profile_updates_units_and_computes_correct_mass(
        self
    ):

        cosmology = mock_cosmology.MockCosmology(kpc_per_arcsec=2.0)

        # arcsec -> arcsec

        sis_arcsec = al.mass_profiles.SphericalIsothermal(
            centre=(al.Length(0.0, "arcsec"), al.Length(0.0, "arcsec")),
            einstein_radius=al.Length(2.0, "arcsec"),
        )

        radius = al.Length(2.0, "arcsec")
        mass = sis_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert math.pi * sis_arcsec.einstein_radius * radius == pytest.approx(
            mass, 1e-3
        )

        # arcsec -> kpc

        radius = al.Length(2.0, "kpc")
        mass = sis_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert math.pi * sis_arcsec.einstein_radius * 1.0 == pytest.approx(mass, 1e-3)

        # 2.0 arcsec = 4.0 kpc, same masses.

        radius = al.Length(2.0, "arcsec")
        mass_arcsec = sis_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        radius = al.Length(4.0, "kpc")
        mass_kpc = sis_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert mass_arcsec == mass_kpc

        # kpc -> kpc

        sis_kpc = al.mass_profiles.SphericalIsothermal(
            centre=(al.Length(0.0, "kpc"), al.Length(0.0, "kpc")),
            einstein_radius=al.Length(2.0, "kpc"),
        )

        radius = al.Length(2.0, "kpc")
        mass = sis_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert math.pi * sis_kpc.einstein_radius * radius == pytest.approx(mass, 1e-3)

        # kpc -> arcsec

        radius = al.Length(2.0, "arcsec")
        mass = sis_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert 2.0 * math.pi * sis_kpc.einstein_radius * radius == pytest.approx(
            mass, 1e-3
        )

        # 2.0 arcsec = 4.0 kpc, same masses.

        radius = al.Length(2.0, "arcsec")
        mass_arcsec = sis_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        radius = al.Length(4.0, "kpc")
        mass_kpc = sis_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert mass_arcsec == mass_kpc

    def test__mass_units_conversions__multiplies_by_critical_surface_density_factor(
        self
    ):

        cosmology = mock_cosmology.MockCosmology(critical_surface_density=2.0)

        sis = al.mass_profiles.SphericalIsothermal(einstein_radius=2.0)
        radius = al.Length(2.0, "arcsec")

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="solMass",
            cosmology=cosmology,
        )
        assert 2.0 * math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="solMass",
            cosmology=cosmology,
        )
        assert 2.0 * math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)


class TestMassWithinEllipse(object):
    def test__mass_in_angular_units__singular_isothermal_sphere__compare_circle_and_ellipse(
        self
    ):

        sis = al.mass_profiles.SphericalIsothermal(einstein_radius=2.0)

        radius = al.Length(2.0)
        mass_circle = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        mass_ellipse = sis.mass_within_ellipse_in_units(
            major_axis=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert mass_circle == mass_ellipse

        sie = al.mass_profiles.EllipticalIsothermal(
            einstein_radius=2.0, axis_ratio=0.5, phi=0.0
        )
        radius = al.Length(2.0)
        mass_circle = sie.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        mass_ellipse = sie.mass_within_ellipse_in_units(
            major_axis=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert mass_circle == mass_ellipse * 2.0

    def test__mass_in_angular_units__singular_isothermal_ellipsoid__compare_to_grid(
        self
    ):

        sie = al.mass_profiles.EllipticalIsothermal(
            einstein_radius=2.0, axis_ratio=0.5, phi=0.0
        )

        radius = al.Length(0.5)

        mass_grid = mass_within_radius_of_profile_from_grid_calculation(
            radius=radius, profile=sie
        )

        mass = sie.mass_within_ellipse_in_units(
            major_axis=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )

        # Large errors required due to cusp at center of SIE - can get to errors of 0.01 for a 400 x 400 grid.
        assert mass_grid == pytest.approx(mass, 0.1)

    def test__radius_units_conversions__mass_profile_updates_units_and_computes_correct_mass(
        self
    ):

        cosmology = mock_cosmology.MockCosmology(kpc_per_arcsec=2.0)

        # arcsec -> arcsec

        sie_arcsec = al.mass_profiles.SphericalIsothermal(
            centre=(al.Length(0.0, "arcsec"), al.Length(0.0, "arcsec")),
            einstein_radius=al.Length(2.0, "arcsec"),
        )

        major_axis = al.Length(0.5, "arcsec")

        mass_grid = mass_within_radius_of_profile_from_grid_calculation(
            radius=major_axis, profile=sie_arcsec
        )

        mass = sie_arcsec.mass_within_ellipse_in_units(
            major_axis=major_axis,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert mass_grid == pytest.approx(mass, 0.1)

        # arcsec -> kpc

        major_axis = al.Length(0.5, "kpc")
        mass = sie_arcsec.mass_within_ellipse_in_units(
            major_axis=major_axis,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert 0.5 * mass_grid == pytest.approx(mass, 0.1)

        # 2.0 arcsec = 4.0 kpc, same masses.

        radius = al.Length(2.0, "arcsec")
        mass_arcsec = sie_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        radius = al.Length(4.0, "kpc")
        mass_kpc = sie_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert mass_arcsec == mass_kpc

        # kpc -> kpc

        sie_kpc = al.mass_profiles.SphericalIsothermal(
            centre=(al.Length(0.0, "kpc"), al.Length(0.0, "kpc")),
            einstein_radius=al.Length(2.0, "kpc"),
        )

        major_axis = al.Length(0.5, "kpc")
        mass = sie_kpc.mass_within_ellipse_in_units(
            major_axis=major_axis,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert mass_grid == pytest.approx(mass, 0.1)

        # kpc -> arcsec

        major_axis = al.Length(0.5, "arcsec")
        mass = sie_kpc.mass_within_ellipse_in_units(
            major_axis=major_axis,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert 2.0 * mass_grid == pytest.approx(mass, 0.1)

        # 2.0 arcsec = 4.0 kpc, same masses.

        radius = al.Length(2.0, "arcsec")
        mass_arcsec = sie_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        radius = al.Length(4.0, "kpc")
        mass_kpc = sie_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )
        assert mass_arcsec == mass_kpc

    def test__mass_unit_conversions__compare_to_grid__mutliplies_by_critical_surface_density(
        self
    ):

        cosmology = mock_cosmology.MockCosmology(critical_surface_density=2.0)

        sie = al.mass_profiles.EllipticalIsothermal(
            einstein_radius=2.0, axis_ratio=0.5, phi=0.0
        )

        radius = al.Length(2.0, "arcsec")

        mass_grid = mass_within_radius_of_profile_from_grid_calculation(
            radius=radius, profile=sie
        )
        mass = sie.mass_within_ellipse_in_units(
            major_axis=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )

        # Large errors required due to cusp at center of SIE - can get to errors of 0.01 for a 400 x 400 grid.
        assert mass_grid == pytest.approx(radius * sie.axis_ratio * mass, 0.1)

        critical_surface_density = al.MassOverLength2(2.0, "arcsec", "solMass")
        mass = sie.mass_within_ellipse_in_units(
            major_axis=radius,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="solMass",
            cosmology=cosmology,
        )

        # Large errors required due to cusp at center of SIE - can get to errors of 0.01 for a 400 x 400 grid.
        assert mass_grid == pytest.approx(0.5 * radius * sie.axis_ratio * mass, 0.1)


class TestDensityBetweenAnnuli(object):
    def test__circular_annuli__sis__analyic_density_agrees(self):

        cosmology = mock_cosmology.MockCosmology(
            kpc_per_arcsec=2.0, critical_surface_density=2.0
        )

        einstein_radius = 1.0
        sis_arcsec = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=einstein_radius
        )

        inner_annuli_radius = al.Length(2.0, "arcsec")
        outer_annuli_radius = al.Length(3.0, "arcsec")

        inner_mass = math.pi * einstein_radius * inner_annuli_radius
        outer_mass = math.pi * einstein_radius * outer_annuli_radius

        density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
            inner_annuli_radius=inner_annuli_radius,
            outer_annuli_radius=outer_annuli_radius,
            unit_length="arcsec",
            unit_mass="angular",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        annuli_area = (np.pi * outer_annuli_radius ** 2.0) - (
            np.pi * inner_annuli_radius ** 2.0
        )

        assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
            density_between_annuli, 1e-4
        )

        density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
            inner_annuli_radius=inner_annuli_radius,
            outer_annuli_radius=outer_annuli_radius,
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        annuli_area = (np.pi * outer_annuli_radius ** 2.0) - (
            np.pi * inner_annuli_radius ** 2.0
        )

        assert (2.0 * outer_mass - 2.0 * inner_mass) / annuli_area == pytest.approx(
            density_between_annuli, 1e-4
        )

        density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
            inner_annuli_radius=inner_annuli_radius,
            outer_annuli_radius=outer_annuli_radius,
            unit_length="kpc",
            unit_mass="angular",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        inner_mass = math.pi * 2.0 * einstein_radius * inner_annuli_radius
        outer_mass = math.pi * 2.0 * einstein_radius * outer_annuli_radius

        annuli_area = (np.pi * 2.0 * outer_annuli_radius ** 2.0) - (
            np.pi * 2.0 * inner_annuli_radius ** 2.0
        )

        assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
            density_between_annuli, 1e-4
        )

    def test__circular_annuli__nfw_profile__compare_to_manual_mass(self):

        cosmology = mock_cosmology.MockCosmology(
            kpc_per_arcsec=2.0, critical_surface_density=2.0
        )

        nfw = al.mass_profiles.EllipticalNFW(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, kappa_s=1.0
        )

        inner_mass = nfw.mass_within_circle_in_units(
            radius=al.Length(1.0),
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )

        outer_mass = nfw.mass_within_circle_in_units(
            radius=al.Length(2.0),
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )

        density_between_annuli = nfw.density_between_circular_annuli_in_angular_units(
            inner_annuli_radius=al.Length(1.0),
            outer_annuli_radius=al.Length(2.0),
            unit_length="arcsec",
            unit_mass="angular",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        annuli_area = (np.pi * 2.0 ** 2.0) - (np.pi * 1.0 ** 2.0)

        assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
            density_between_annuli, 1e-4
        )


class TestDeflectionsViaPotential(object):
    def test__compare_sis_deflections_via_potential_and_calculation(self):

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.05
        )

        deflections_via_calculation = sis.deflections_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        deflections_via_potential = sis.deflections_via_potential_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        mean_error = np.mean(deflections_via_potential - deflections_via_calculation)

        assert mean_error < 1e-4

    def test__compare_sie_at_phi_45__deflections_via_potential_and_calculation(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=45.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.05
        )

        deflections_via_calculation = sie.deflections_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        deflections_via_potential = sie.deflections_via_potential_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        mean_error = np.mean(deflections_via_potential - deflections_via_calculation)

        assert mean_error < 1e-4

    def test__compare_sie_at_phi_0__deflections_via_potential_and_calculation(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.05
        )

        deflections_via_calculation = sie.deflections_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        deflections_via_potential = sie.deflections_via_potential_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        mean_error = np.mean(deflections_via_potential - deflections_via_calculation)

        assert mean_error < 1e-4

    def test(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.05, sub_grid_size=2
        )

        deflections_binned_reg_grid = sie.deflections_via_potential_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        deflections_sub_grid = sie.deflections_via_potential_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        pixel_1_reg_grid = deflections_binned_reg_grid[0]

        pixel_1_from_av_sub_grid = (
            deflections_sub_grid[0]
            + deflections_sub_grid[1]
            + deflections_sub_grid[2]
            + deflections_sub_grid[3]
        ) / 4

        assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

        pixel_10000_reg_grid = deflections_binned_reg_grid[99]

        pixel_10000_from_av_sub_grid = (
            deflections_sub_grid[399]
            + deflections_sub_grid[398]
            + deflections_sub_grid[397]
            + deflections_sub_grid[396]
        ) / 4

        assert pixel_10000_reg_grid == pytest.approx(pixel_10000_from_av_sub_grid, 1e-4)


class TestJacobianandMagnification(object):
    def test__jacobian_components(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05
        )

        jacobian = sie.lensing_jacobian_from_grid(grid=grid, return_in_2d=False)

        A_12 = jacobian[0, 1]
        A_21 = jacobian[1, 0]

        mean_error = np.mean(A_12 - A_21)

        assert mean_error < 1e-4

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
        )

        jacobian = sie.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        A_12 = jacobian[0, 1]
        A_21 = jacobian[1, 0]

        mean_error = np.mean(A_12 - A_21)

        assert mean_error < 1e-4

    def test__compare_magnification_from_eigen_values_and_from_determinant(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=1
        )

        magnification_via_determinant = sie.magnification_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        tangential_eigen_value = sie.tangential_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        radal_eigen_value = sie.radial_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        magnification_via_eigen_values = 1 / (
            tangential_eigen_value * radal_eigen_value
        )

        mean_error = np.mean(
            magnification_via_determinant - magnification_via_eigen_values
        )

        assert mean_error < 1e-4

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
        )

        magnification_via_determinant = sie.magnification_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        tangential_eigen_value = sie.tangential_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        radal_eigen_value = sie.radial_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        magnification_via_eigen_values = 1 / (
            tangential_eigen_value * radal_eigen_value
        )

        mean_error = np.mean(
            magnification_via_determinant - magnification_via_eigen_values
        )

        assert mean_error < 1e-4

    def test__compare_magnification_from_determinant_and_from_convergence_and_shear(
        self
    ):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=1
        )

        magnification_via_determinant = sie.magnification_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        convergence = sie.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        shear = sie.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        magnification_via_convergence_and_shear = 1 / (
            (1 - convergence) ** 2 - shear ** 2
        )

        mean_error = np.mean(
            magnification_via_determinant - magnification_via_convergence_and_shear
        )

        assert mean_error < 1e-4

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
        )

        magnification_via_determinant = sie.magnification_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        convergence = sie.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        shear = sie.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        magnification_via_convergence_and_shear = 1 / (
            (1 - convergence) ** 2 - shear ** 2
        )

        mean_error = np.mean(
            magnification_via_determinant - magnification_via_convergence_and_shear
        )

        assert mean_error < 1e-4

    def test__jacobian(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.05, sub_grid_size=2
        )

        jacobian_binned_reg_grid = sie.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )
        a11_binned_reg_grid = jacobian_binned_reg_grid[0, 0]

        jacobian_sub_grid = sie.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )
        a11_sub_grid = jacobian_sub_grid[0, 0]

        pixel_1_reg_grid = a11_binned_reg_grid[0]
        pixel_1_from_av_sub_grid = (
            a11_sub_grid[0] + a11_sub_grid[1] + a11_sub_grid[2] + a11_sub_grid[3]
        ) / 4

        assert jacobian_binned_reg_grid.shape == (2, 2, 100)
        assert jacobian_sub_grid.shape == (2, 2, 400)
        assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

        pixel_10000_reg_grid = a11_binned_reg_grid[99]

        pixel_10000_from_av_sub_grid = (
            a11_sub_grid[399]
            + a11_sub_grid[398]
            + a11_sub_grid[397]
            + a11_sub_grid[396]
        ) / 4

        assert pixel_10000_reg_grid == pytest.approx(pixel_10000_from_av_sub_grid, 1e-4)

    def test__shear(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.05, sub_grid_size=2
        )

        shear_binned_reg_grid = sie.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        shear_sub_grid = sie.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        pixel_1_reg_grid = shear_binned_reg_grid[0]
        pixel_1_from_av_sub_grid = (
            shear_sub_grid[0]
            + shear_sub_grid[1]
            + shear_sub_grid[2]
            + shear_sub_grid[3]
        ) / 4

        assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

        pixel_10000_reg_grid = shear_binned_reg_grid[99]

        pixel_10000_from_av_sub_grid = (
            shear_sub_grid[399]
            + shear_sub_grid[398]
            + shear_sub_grid[397]
            + shear_sub_grid[396]
        ) / 4

        assert pixel_10000_reg_grid == pytest.approx(pixel_10000_from_av_sub_grid, 1e-4)

    def test__lambda_t(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.05, sub_grid_size=2
        )

        lambda_t_binned_reg_grid = sie.tangential_eigen_value_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        lambda_t_sub_grid = sie.tangential_eigen_value_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        pixel_1_reg_grid = lambda_t_binned_reg_grid[0]
        pixel_1_from_av_sub_grid = (
            lambda_t_sub_grid[0]
            + lambda_t_sub_grid[1]
            + lambda_t_sub_grid[2]
            + lambda_t_sub_grid[3]
        ) / 4

        assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

        pixel_10000_reg_grid = lambda_t_binned_reg_grid[99]

        pixel_10000_from_av_sub_grid = (
            lambda_t_sub_grid[399]
            + lambda_t_sub_grid[398]
            + lambda_t_sub_grid[397]
            + lambda_t_sub_grid[396]
        ) / 4

        assert pixel_10000_reg_grid == pytest.approx(pixel_10000_from_av_sub_grid, 1e-4)

    def test__lambda_r_from_grid(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
        )

        lambda_r_binned_reg_grid = sie.radial_eigen_value_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        lambda_r_sub_grid = sie.radial_eigen_value_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        pixel_1_reg_grid = lambda_r_binned_reg_grid[0]
        pixel_1_from_av_sub_grid = (
            lambda_r_sub_grid[0]
            + lambda_r_sub_grid[1]
            + lambda_r_sub_grid[2]
            + lambda_r_sub_grid[3]
        ) / 4

        assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

        pixel_10000_reg_grid = lambda_r_binned_reg_grid[99]

        pixel_10000_from_av_sub_grid = (
            lambda_r_sub_grid[399]
            + lambda_r_sub_grid[398]
            + lambda_r_sub_grid[397]
            + lambda_r_sub_grid[396]
        ) / 4

        assert pixel_10000_reg_grid == pytest.approx(pixel_10000_from_av_sub_grid, 1e-4)


def critical_curve_via_magnification_from_mass_profile_and_grid(mass_profile, grid):
    magnification_2d = mass_profile.magnification_from_grid(
        grid=grid, return_in_2d=True, return_binned=False
    )

    inverse_magnification_2d = 1 / magnification_2d

    critical_curves_indices = measure.find_contours(inverse_magnification_2d, 0)

    no_critical_curves = len(critical_curves_indices)
    contours = []
    critical_curves = []

    for jj in np.arange(no_critical_curves):

        contours.append(critical_curves_indices[jj])
        contour_x, contour_y = contours[jj].T
        pixel_coord = np.stack((contour_x, contour_y), axis=-1)

        critical_curve = grid.marching_squares_grid_pixels_to_grid_arcsec(
            grid_pixels=pixel_coord, shape=magnification_2d.shape
        )

        critical_curves.append(critical_curve)

    return critical_curves


def caustics_via_magnification_from_mass_profile_and_grid(mass_profile, grid):

    caustics = []

    critical_curves = critical_curve_via_magnification_from_mass_profile_and_grid(
        mass_profile=mass_profile, grid=grid
    )

    for i in range(len(critical_curves)):

        critical_curve = critical_curves[i]

        deflections_1d = mass_profile.deflections_from_grid(
            grid=critical_curve, return_in_2d=False, return_binned=False
        )

        caustic = critical_curve - deflections_1d

        caustics.append(caustic)

    return caustics


class TestConvergenceViajacobian(object):
    def test__compare_sis_convergence_via_jacobian_and_calculation(self):

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.05
        )

        convergence_via_calculation = sis.convergence_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        convergence_via_jacobian = sis.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        mean_error = np.mean(convergence_via_jacobian - convergence_via_calculation)

        assert convergence_via_jacobian.shape == (20, 20)
        assert mean_error < 1e-1

        convergence_via_calculation = sis.convergence_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        convergence_via_jacobian = sis.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        mean_error = np.mean(convergence_via_jacobian - convergence_via_calculation)

        assert convergence_via_jacobian.shape == (400,)
        assert mean_error < 1e-1

    def test__compare_sie_at_phi_45__convergence_via_jacobian_and_calculation(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=45.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.05
        )

        convergence_via_calculation = sie.convergence_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        convergence_via_jacobian = sie.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        mean_error = np.mean(convergence_via_jacobian - convergence_via_calculation)

        assert mean_error < 1e-1

    def test__convergence(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.05, sub_grid_size=2
        )

        convergence_binned_reg_grid = sie.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        convergence_sub_grid = sie.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        pixel_1_reg_grid = convergence_binned_reg_grid[0]
        pixel_1_from_av_sub_grid = (
            convergence_sub_grid[0]
            + convergence_sub_grid[1]
            + convergence_sub_grid[2]
            + convergence_sub_grid[3]
        ) / 4

        assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

        pixel_10000_reg_grid = convergence_binned_reg_grid[99]

        pixel_10000_from_av_sub_grid = (
            convergence_sub_grid[399]
            + convergence_sub_grid[398]
            + convergence_sub_grid[397]
            + convergence_sub_grid[396]
        ) / 4

        assert pixel_10000_reg_grid == pytest.approx(pixel_10000_from_av_sub_grid, 1e-4)

        convergence_via_calculation = sie.convergence_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        convergence_via_jacobian = sie.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        mean_error = np.mean(convergence_via_jacobian - convergence_via_calculation)

        assert convergence_via_jacobian.shape == (400,)
        assert mean_error < 1e-1


class TestCriticalCurvesandCaustics(object):
    def test_compare_magnification_from_determinant_and_from_convergence_and_shear(
        self
    ):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), phi=0.0, axis_ratio=0.8, einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
        )

        magnification_via_determinant = sie.magnification_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        convergence = sie.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        shear = sie.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        magnification_via_convergence_and_shear = 1 / (
            (1 - convergence) ** 2 - shear ** 2
        )

        mean_error = np.mean(
            magnification_via_determinant - magnification_via_convergence_and_shear
        )

        assert mean_error < 1e-2

    def test__tangential_critical_curve_radii__spherical_isothermal(self):

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=2
        )

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        tangential_critical_curve = critical_curves[0]

        x_critical_tangential, y_critical_tangential = (
            tangential_critical_curve[:, 1],
            tangential_critical_curve[:, 0],
        )

        assert np.mean(
            x_critical_tangential ** 2 + y_critical_tangential ** 2
        ) == pytest.approx(sis.einstein_radius ** 2, 5e-1)

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.5, sub_grid_size=4
        )

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        tangential_critical_curve = critical_curves[0]

        x_critical_tangential, y_critical_tangential = (
            tangential_critical_curve[:, 1],
            tangential_critical_curve[:, 0],
        )

        assert np.mean(
            x_critical_tangential ** 2 + y_critical_tangential ** 2
        ) == pytest.approx(sis.einstein_radius ** 2, 5e-1)

    def test__tangential_critical_curve_centres__spherical_isothermal(self):

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=1
        )

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        tangential_critical_curve = critical_curves[0]

        y_centre = np.mean(tangential_critical_curve[:, 0])
        x_centre = np.mean(tangential_critical_curve[:, 1])

        assert -0.03 < y_centre < 0.03
        assert -0.03 < x_centre < 0.03

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=4
        )

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        tangential_critical_curve = critical_curves[0]

        y_centre = np.mean(tangential_critical_curve[:, 0])
        x_centre = np.mean(tangential_critical_curve[:, 1])

        assert -0.01 < y_centre < 0.01
        assert -0.01 < x_centre < 0.01

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.5, 1.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(60, 60), pixel_scale=0.25, sub_grid_size=1
        )

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        tangential_critical_curve = critical_curves[0]

        y_centre = np.mean(tangential_critical_curve[:, 0])
        x_centre = np.mean(tangential_critical_curve[:, 1])

        assert 0.47 < y_centre < 0.53
        assert 0.97 < x_centre < 1.03

    def test__radial_critical_curve_centres__spherical_isothermal(self):

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=1
        )

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        radial_critical_curve = critical_curves[1]

        y_centre = np.mean(radial_critical_curve[:, 0])
        x_centre = np.mean(radial_critical_curve[:, 1])

        assert -0.05 < y_centre < 0.05
        assert -0.05 < x_centre < 0.05

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=4
        )

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        radial_critical_curve = critical_curves[1]

        y_centre = np.mean(radial_critical_curve[:, 0])
        x_centre = np.mean(radial_critical_curve[:, 1])

        assert -0.01 < y_centre < 0.01
        assert -0.01 < x_centre < 0.01

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.5, 1.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(60, 60), pixel_scale=0.25, sub_grid_size=1
        )

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        radial_critical_curve = critical_curves[1]

        y_centre = np.mean(radial_critical_curve[:, 0])
        x_centre = np.mean(radial_critical_curve[:, 1])

        assert 0.45 < y_centre < 0.55
        assert 0.95 < x_centre < 1.05

    def test__tangential_caustic_centres__spherical_isothermal(self):

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=1
        )

        caustics = sis.caustics_from_grid(grid=grid)

        tangential_caustic = caustics[0]

        y_centre = np.mean(tangential_caustic[:, 0])
        x_centre = np.mean(tangential_caustic[:, 1])

        assert -0.03 < y_centre < 0.03
        assert -0.03 < x_centre < 0.03

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=4
        )

        caustics = sis.caustics_from_grid(grid=grid)

        tangential_caustic = caustics[0]

        y_centre = np.mean(tangential_caustic[:, 0])
        x_centre = np.mean(tangential_caustic[:, 1])

        assert -0.01 < y_centre < 0.01
        assert -0.01 < x_centre < 0.01

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.5, 1.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(60, 60), pixel_scale=0.25, sub_grid_size=1
        )

        caustics = sis.caustics_from_grid(grid=grid)

        tangential_caustic = caustics[0]

        y_centre = np.mean(tangential_caustic[:, 0])
        x_centre = np.mean(tangential_caustic[:, 1])

        assert 0.47 < y_centre < 0.53
        assert 0.97 < x_centre < 1.03

    def test__radial_caustics_radii__spherical_isothermal(self):

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.5, sub_grid_size=4
        )

        caustics = sis.caustics_from_grid(grid=grid)

        caustic_radial = caustics[1]

        x_caustic_radial, y_caustic_radial = (
            caustic_radial[:, 1],
            caustic_radial[:, 0],
        )

        assert np.mean(x_caustic_radial ** 2 + y_caustic_radial ** 2) == pytest.approx(
            sis.einstein_radius ** 2, 5e-1
        )

    def test__radial_caustic_centres__spherical_isothermal(self):

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=1
        )

        caustics = sis.caustics_from_grid(grid=grid)

        radial_caustic = caustics[1]

        y_centre = np.mean(radial_caustic[:, 0])
        x_centre = np.mean(radial_caustic[:, 1])

        assert -0.2 < y_centre < 0.2
        assert -0.2 < x_centre < 0.2

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25, sub_grid_size=4
        )

        caustics = sis.caustics_from_grid(grid=grid)

        radial_caustic = caustics[1]

        y_centre = np.mean(radial_caustic[:, 0])
        x_centre = np.mean(radial_caustic[:, 1])

        assert -0.09 < y_centre < 0.09
        assert -0.09 < x_centre < 0.09

        sis = al.mass_profiles.SphericalIsothermal(
            centre=(0.5, 1.0), einstein_radius=2.0
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(60, 60), pixel_scale=0.25, sub_grid_size=1
        )

        caustics = sis.caustics_from_grid(grid=grid)

        radial_caustic = caustics[1]

        y_centre = np.mean(radial_caustic[:, 0])
        x_centre = np.mean(radial_caustic[:, 1])

        assert 0.3 < y_centre < 0.7
        assert 0.8 < x_centre < 1.2

    def test__compare_tangential_critical_curves_from_magnification_and_lamda_t(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2, axis_ratio=0.8, phi=40
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25
        )

        tangential_critical_curve_from_magnification = critical_curve_via_magnification_from_mass_profile_and_grid(
            mass_profile=sie, grid=grid
        )[
            0
        ]

        tangential_critical_curve_from_lambda_t = sie.tangential_critical_curve_from_grid(
            grid=grid
        )

        assert tangential_critical_curve_from_lambda_t == pytest.approx(
            tangential_critical_curve_from_magnification, 5e-1
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.5, sub_grid_size=2
        )

        tangential_critical_curve_from_magnification = critical_curve_via_magnification_from_mass_profile_and_grid(
            mass_profile=sie, grid=grid
        )[
            0
        ]

        tangential_critical_curve_from_lambda_t = sie.tangential_critical_curve_from_grid(
            grid=grid
        )

        assert tangential_critical_curve_from_lambda_t == pytest.approx(
            tangential_critical_curve_from_magnification, 5e-1
        )

    def test__compare_radial_critical_curves_from_magnification_and_lamda_r(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2, axis_ratio=0.8, phi=40
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
        )

        critical_curve_radial_from_magnification = critical_curve_via_magnification_from_mass_profile_and_grid(
            mass_profile=sie, grid=grid
        )[
            1
        ]

        critical_curve_radial_from_lambda_t = sie.radial_critical_curve_from_grid(
            grid=grid
        )

        assert sum(critical_curve_radial_from_magnification) == pytest.approx(
            sum(critical_curve_radial_from_lambda_t), 5e-1
        )

    def test__compare_tangential_caustic_from_magnification_and_lambda_t(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2, axis_ratio=0.8, phi=40
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.25
        )

        tangential_caustic_from_magnification = caustics_via_magnification_from_mass_profile_and_grid(
            mass_profile=sie, grid=grid
        )[
            0
        ]

        tangential_caustic_from_lambda_t = sie.tangential_caustic_from_grid(grid=grid)

        assert sum(tangential_caustic_from_lambda_t) == pytest.approx(
            sum(tangential_caustic_from_magnification), 5e-1
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=0.5, sub_grid_size=2
        )

        tangential_caustic_from_magnification = caustics_via_magnification_from_mass_profile_and_grid(
            mass_profile=sie, grid=grid
        )[
            0
        ]

        tangential_caustic_from_lambda_t = sie.tangential_caustic_from_grid(grid=grid)

        assert sum(tangential_caustic_from_lambda_t) == pytest.approx(
            sum(tangential_caustic_from_magnification), 5e-1
        )

    def test__compare_radial_caustic_from_magnification_and_lambda_r__grid(self):

        sie = al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2, axis_ratio=0.8, phi=40
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05
        )

        caustic_radial_from_magnification = caustics_via_magnification_from_mass_profile_and_grid(
            mass_profile=sie, grid=grid
        )[
            1
        ]

        caustic_radial_from_lambda_r = sie.caustics_from_grid(grid=grid)[1]

        assert sum(caustic_radial_from_lambda_r) == pytest.approx(
            sum(caustic_radial_from_magnification), 7e-1
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
        )

        caustic_radial_from_magnification = caustics_via_magnification_from_mass_profile_and_grid(
            mass_profile=sie, grid=grid
        )[
            1
        ]

        caustic_radial_from_lambda_r = sie.caustics_from_grid(grid=grid)[1]

        assert sum(caustic_radial_from_lambda_r) == pytest.approx(
            sum(caustic_radial_from_magnification), 5e-1
        )



import autofit as af
import autolens as al

# In this pipeline, we'll perform an analysis which fits an image with  no lens light, and a source galaxy using a
# parametric light profile, using a power-law mass profile. The pipeline follows on from the initialize pipeline
# ''pipelines/no_lens_light/initialize/lens_sie__source_sersic_from_init.py'.

# The pipeline is one phase, as follows:

# Phase 1:

# Description: Fits the lens mass model as a power-law, using a parametric Sersic light profile for the source.
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: no_lens_light/initialize/lens_sie__source_sersic_from_init.py
# Prior Passing: None
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations.


def make_pipeline(
    pipeline_settings,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    pixel_scale_interpolation_grid=None,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_power_law__lens_power_law__source_sersic"

    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings(
        include_shear=pipeline_settings.include_shear
    )

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy mass using the EllipticalIsothermal fit of the previous pipeline, and
    #    source galaxy of the previous pipeline.

    class LensSourcePhase(al.PhaseImaging):
        def customize_priors(self, results):

            ### Lens Mass, SIE -> PL ###

            self.galaxies.lens.mass.centre = results.from_phase(
                "phase_1__lens_sie__source_sersic"
            ).variable.galaxies.lens.mass.centre

            self.galaxies.lens.mass.axis_ratio = results.from_phase(
                "phase_1__lens_sie__source_sersic"
            ).variable.galaxies.lens.mass.axis_ratio

            self.galaxies.lens.mass.phi = results.from_phase(
                "phase_1__lens_sie__source_sersic"
            ).variable.galaxies.lens.mass.phi

            self.galaxies.lens.mass.einstein_radius = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_absolute(a=0.3)
                .galaxies.lens.mass.einstein_radius
            )

            ### Lens Shear, Shear -> Shear ###

            if pipeline_settings.include_shear:

                self.galaxies.lens.shear = results.from_phase(
                    "phase_1__lens_sie__source_sersic"
                ).variable.galaxies.lens.shear

            ### Source Light, Sersic -> Sersic ###

            self.galaxies.source = results.from_phase(
                "phase_1__lens_sie__source_sersic"
            ).variable.galaxies.source

    phase1 = LensSourcePhase(
        phase_name="phase_1__lens_power_law__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                mass=al.mass_profiles.EllipticalPowerLaw,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source, light=al.light_profiles.EllipticalSersic
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 75
    phase1.optimizer.sampling_efficiency = 0.2

    return al.PipelineImaging(pipeline_name, phase1)



import autofit as af
import autolens as al

import os

# This tool allows one to make simulated data-sets of strong lenses, which can be used to test example pipelines and
# investigate strong lens modeling on data-sets where the 'true' answer is known.

# The 'data name' is the name of the data folder and 'data_name' the folder the data is stored in, e.g:

# The image will be output as '/workspace/data/data_type/data_name/image.fits'.
# The noise-map will be output as '/workspace/data/data_type/data_name/lens_name/noise_map.fits'.
# The psf will be output as '/workspace/data/data_type/data_name/psf.fits'.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# (these files are already in the workspace and are remade running this script)
data_type = "example"
data_name = "lens_sersic_sie__source_sersic"

# Create the path where the data will be output, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

# The pixel scale of the image to be simulated
pixel_scale = 0.1

# Simulate a simple Gaussian PSF for the image.
psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=pixel_scale)

# Setup the image-plane al.ogrid of the CCD array which will be used for generating the image of the
# simulated strong lens. The sub-grid size of 20x20 ensures we fully resolve the central regions of the lens and source
# galaxy light.
grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=pixel_scale, sub_grid_size=16
)

# Setup the lens galaxy's light (elliptical Sersic), mass (SIE+Shear) and source galaxy light (elliptical Sersic) for
# this simulated lens.
lens_galaxy = al.Galaxy(
    redshift=0.5,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.9,
        phi=45.0,
        intensity=1.0,
        effective_radius=0.8,
        sersic_index=4.0,
    ),
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.3,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)


# Use these galaxies to setup a tracer, which will generate the image for the simulated CCD instrument.
tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

# Lets look at the tracer's image - this is the image we'll be simulating.
al.ray_tracing_plotters.plot_profile_image(tracer=tracer, grid=grid)

# Simulate the CCD instrument, remembering that we use a special image which ensures edge-effects don't
# degrade our modeling of the telescope optics (e.g. the PSF convolution).
ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)

# Lets plot the simulated CCD instrument before we output it to files.
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Finally, lets output our simulated instrument to the data path as .fits files.
al.output_ccd_data_to_fits(
    ccd_data=ccd_data,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)


# ####################################### OTHER EXAMPLE IMAGES #####################################

data_type = "example"
data_name = "lens_sersic_sie__source_sersic__2"


data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.7,
        phi=80.0,
        intensity=0.8,
        effective_radius=1.3,
        sersic_index=2.5,
    ),
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.3, axis_ratio=0.6, phi=30.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.02, phi=145.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(-0.2, -0.3),
        axis_ratio=0.9,
        phi=10.0,
        intensity=0.2,
        effective_radius=1.5,
        sersic_index=2.0,
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)


# Finally, lets output our simulated instrument to the data path as .fits files.
al.output_ccd_data_to_fits(
    ccd_data=ccd_data,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)

####################################### OTHER EXAMPLE IMAGES #####################################

data_type = "example"
data_name = "lens_sie__source_sersic"

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.3,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)


# Finally, lets output our simulated instrument to the data path as .fits files.
al.output_ccd_data_to_fits(
    ccd_data=ccd_data,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)

####################################### OTHER EXAMPLE IMAGES #####################################

data_type = "example"
data_name = "lens_sie__source_sersic__2"

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.3, axis_ratio=0.8, phi=60.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.3, -0.4),
        axis_ratio=0.6,
        phi=40.0,
        intensity=0.2,
        effective_radius=1.2,
        sersic_index=2.0,
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)


# Finally, lets output our simulated instrument to the data path as .fits files.
al.output_ccd_data_to_fits(
    ccd_data=ccd_data,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)

####################################################

data_type = "example"
data_name = "lens_mass__source_sersic_x2"

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.9, phi=90.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy_0 = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.25, 0.15),
        axis_ratio=0.7,
        phi=120.0,
        intensity=0.7,
        effective_radius=0.7,
        sersic_index=1.0,
    ),
)

source_galaxy_1 = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.7, -0.5),
        axis_ratio=0.9,
        phi=60.0,
        intensity=0.2,
        effective_radius=1.6,
        sersic_index=3.0,
    ),
)

tracer = al.Tracer.from_galaxies(
    galaxies=[lens_galaxy, source_galaxy_0, source_galaxy_1]
)

ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)

al.output_ccd_data_to_fits(
    ccd_data=ccd_data,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)

####################################################

data_type = "example"
data_name = "lens_bulge_disk_sie__source_sersic"

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    bulge=al.light_profiles.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.9,
        phi=45.0,
        intensity=0.3,
        effective_radius=0.6,
        sersic_index=3.0,
    ),
    disk=al.light_profiles.EllipticalExponential(
        centre=(0.0, 0.0), axis_ratio=0.7, phi=30.0, intensity=0.2, effective_radius=1.6
    ),
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.3,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)

# Lets plot the simulated CCD instrument before we output it to files.
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

al.output_ccd_data_to_fits(
    ccd_data=ccd_data,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)


####################################### OTHER EXAMPLE IMAGES #####################################

data_type = "example"
data_name = "lens_sie__source_sersic__intervening_objects"

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    intervene_0=al.light_profiles.SphericalExponential(
        centre=(1.0, 3.5), intensity=0.8, effective_radius=0.5
    ),
    intervene_1=al.light_profiles.SphericalExponential(
        centre=(-2.0, -3.5), intensity=0.5, effective_radius=0.8
    ),
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.3,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)

# Finally, lets output our simulated instrument to the data path as .fits files.
al.output_ccd_data_to_fits(
    ccd_data=ccd_data,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)



import autolens as al
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion.plotters import mapper_plotters

# To begin chapter 4, we'll begin by learning about pixelizations, which we apply to a source-plane to reconstruct a
# source galaxy's light.

# Lets setup a lensed source-plane grid, using a lens galaxy and tracer (our source galaxy doesn't have a light profile,
# as we're going to reconstruct its light using a pixelization).
grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, einstein_radius=1.6
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, al.Galaxy(redshift=1.0)])

source_plane_grid = tracer.traced_grids_of_planes_from_grid(grid=grid)[1]

# Next, lets set up a pixelization using the 'pixelizations' module, which we've imported as 'pix'.

# There are multiple pixelizations available in PyAutoLens. For now, we'll keep it simple and use a uniform
# rectangular grid. As usual, the grid's 'shape' defines its (y,x) dimensions.
rectangular = pix.Rectangular(shape=(25, 25))

# By itself, a pixelization doesn't tell us much. It has no grid of coordinates, no image, and nothing which tells it
# about the lens we're fitting. This information comes when we use the pixelization to set up a 'mapper'. We'll use
# the (traced) source-plane grid to set up this mapper.

mapper = rectangular.mapper_from_grid_and_pixelization_grid(grid=source_plane_grid)

# This mapper is a 'RectangularMapper' - every pixelization generates it owns mapper.
print(type(mapper))

# By plotting our mapper, we now see our pixelization. Its a fairly boring grid of rectangular pixels.
mapper_plotters.plot_rectangular_mapper(
    mapper=mapper,
    should_plot_grid=False,
    title="Fairly Boring Grid of Rectangular Pixels",
)

# However, the mapper does contain lots of interesting information about our pixelization, for example its geometry
# attribute tells us where the pixel centers are located.
print("Rectangular Grid Pixel Centre 1:")
print(mapper.geometry.pixel_centres[0])
print("Rectangular Grid Pixel Centre 2:")
print(mapper.geometry.pixel_centres[1])
print("Rectangular Grid Pixel Centre 3:")
print(mapper.geometry.pixel_centres[2])
print("etc.")

# Infact, we can plot these centre on our grid - to make it look slightly less boring!
mapper_plotters.plot_rectangular_mapper(
    mapper=mapper, should_plot_grid=False, should_plot_centres=True
)

# The mapper also has the (source-plane) grid that we passed when we set it up. Lets check they're the same.
print("Source Grid Pixel 1")
print(source_plane_grid[0])
print(mapper.grid[0])
print("Source Grid Pixel 2")
print(source_plane_grid[1])
print(mapper.grid[1])
print("etc.")

# We can over-lay the grid on top. Its starting to look a bit less boring now!
mapper_plotters.plot_rectangular_mapper(
    mapper=mapper, should_plot_centres=True, should_plot_grid=True
)


# Finally, the mapper and its geometry has lots more information about the pixelization, for example, the arc-second
# size and dimensions.
print(mapper.geometry.shape_arcsec)
print(mapper.geometry.arc_second_maxima)
print(mapper.geometry.arc_second_minima)

# And with that, we're done. This was a relatively gentle overview of pixelizations, but one that
# was hopefully easy to follow. Think about the following questions before moving on to the next tutorial:

# 1) Look at how the source-grid coordinates are distributed over the rectangular pixel-grid. Are these points
#    distributed evenly over the rectangular grid's pixels? Do some pixels have a lot more grid-points inside of them?
#    Do some pixels have no grid-points in them?

#  2) The rectangular pixelization's edges are aligned with the most exterior coordinates of the source-grid. This is
#     intentional - why do you think this is?



import autofit as af
import autolens as al

import os

# In this example, we'll generate a phase which fits a simple lens + source plane system. Whilst I would generally
# recommend that you write pipelines when using PyAutoLens, it can be convenient to sometimes perform non-linear
# searches in one phase to get results quickly.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))

# There is a x2 '/../../' because we are in the 'workspace/scripts/examples' folder. If you write your own script \
# in the 'workspace/script' folder you should remove one '../', as shown below.
# path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output papth
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=workspace_path + "output"
)

# It is convinient to specify the lens name as a string, so that if the pipeline is applied to multiple images we \
# don't have to change all of the path entries in the load_ccd_data_from_fits function below.
data_type = "example"
data_name = "lens_sersic_sie__source_sersic"
pixel_scale = 0.1

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

ccd_data = al.load_ccd_data_from_fits(
    image_path=data_path + "/image.fits",
    psf_path=data_path + "/psf.fits",
    noise_map_path=data_path + "/noise_map.fits",
    pixel_scale=pixel_scale,
)

# The phase can be passed a mask, which we setup below as a 3.0" circle.
mask = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0
)

# We can also specify a set of positions, which must be traced within a threshold value or else the mass model is
positions = al.load_positions(positions_path=data_path + "/positions.dat")

# resampled (see howtolens/chapter_2_lens_modeling/tutorial_7_masking_and_positions.ipynb)
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data, mask=mask, positions=positions)

# We're going to model our lens galaxy using a light profile (an elliptical Sersic) and mass profile
# (a singular isothermal ellipsoid). We load these profiles from the 'light_profiles (lp)' and 'mass_profiles (mp)'.

# To setup our model galaxies, we use the 'galaxy_model' module and GalaxyModel class.
# A GalaxyModel represents a galaxy where the parameters of its associated profiles are
# variable and fitted for by the analysis.
lens_galaxy_model = al.GalaxyModel(
    redshift=0.5,
    light=al.light_profiles.EllipticalSersic,
    mass=al.mass_profiles.EllipticalIsothermal,
)

source_galaxy_model = al.GalaxyModel(
    redshift=1.0, light=al.light_profiles.EllipticalSersic
)

# To perform the analysis, we set up a phase using the 'phase' module (imported as 'ph').
# A phase takes our galaxy models and fits their parameters using a non-linear search (in this case, MultiNest).

# The phase folders and phase name mean the output of these run will be in the directory
# 'workspace/output/example/lens_light_and_x1_source/phase_example'
phase = al.PhaseImaging(
    phase_name="phase_example",
    phase_folders=[data_type, data_name],
    galaxies=dict(lens=lens_galaxy_model, source=source_galaxy_model),
    optimizer_class=af.MultiNest,
)

# You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
# non-linear parameter space faster (if you haven't already, checkout the tutorial '' in howtolens/chapter_2).

phase.optimizer.const_efficiency_mode = True
phase.optimizer.n_live_points = 50
phase.optimizer.sampling_efficiency = 0.5

# We run the phase on the image, print the results and plot the fit.
result = phase.run(data=ccd_data)

al.lens_fit_plotters.plot_fit_subplot(fit=result.most_likely_fit)



import os
import numpy as np

# Before reading this script, you should checkout the 'galaxy_fit_surface_density.py' script first, which shows you
# how to simulate a convergence profile and fit it with a galaxy. In this script, we'll do the same thing with
# deflection angles and using multiple galaxies. There a few benefits to fitting deflection angles instead of a surface
# density profile (or gravitational potential):

# 1) In terms of lensing, the deflection-angle map is closest thing to what we *actually* observe when we image and
#    model a strong lens. Thus fitting deflection angle maps is the best way we can compare the results of a lens model
#    to a theoretical quantity.

# 2) As we do in this example, we can simulate our deflecton angle map using multi-plane lens ray-tracing, and thus
#    investigate the impact assuming a single-lens plane has on the inferred lens model.

### AUTOFIT + CONFIG SETUP ###

import autofit as af

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=workspace_path + "output"
)

### AUTOLENS + DATA SETUP ###

import autolens as al

# First, we'll setup the al.ogrid we use to simulate a deflection profile.
pixel_scale = 0.05
image_shape = (250, 250)
grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=image_shape, pixel_scale=pixel_scale, sub_grid_size=4
)

# Now lets create two galaxies, using singular isothermal spheres. We'll put the two galaxies at different redshifts,
# and the second galaxy will be much lower mass as if it is a 'perturber' of the main lens galaxy.
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0),
)

perturber = al.Galaxy(
    redshift=0.2,
    mass=al.mass_profiles.SphericalIsothermal(centre=(0.5, 0.5), einstein_radius=0.1),
)

# We only need the source galaxy to have a redshift - given we're not fitting an image it doens't need a light profile.
source_galaxy = al.Galaxy(redshift=1.0)

# We'll use a tracer to compute our multi-plane deflection angles.
tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, perturber, source_galaxy])

# We'll now extract the deflection angles from the tracer - we will extract the two deflection angle maps (y and x)
# separately.
deflections_y = tracer.deflections_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)[:, :, 0]
deflections_y = al.ScaledSquarePixelArray(array=deflections_y, pixel_scale=pixel_scale)

deflections_x = tracer.deflections_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)[:, :, 1]
deflections_x = al.ScaledSquarePixelArray(array=deflections_x, pixel_scale=pixel_scale)

# Next we create each deflection angle map as its own GalaxyData object. A somewhat arbritary noise-map is required by
# the fit.
noise_map = al.ScaledSquarePixelArray(
    array=0.1 * np.ones(deflections_y.shape), pixel_scale=pixel_scale
)
data_y = al.GalaxyData(
    image=deflections_y, noise_map=noise_map, pixel_scale=pixel_scale
)
data_x = al.GalaxyData(
    image=deflections_x, noise_map=noise_map, pixel_scale=pixel_scale
)

# The fit will use a mask, which we setup like any other fit. Lets use a circular mask of 2.0"
def mask_function_circular(image):
    return al.Mask.circular(
        shape=image.shape, pixel_scale=image.pixel_scale, radius_arcsec=2.5
    )


# Again, we'll use a special phase, called a 'GalaxyFitPhase', to fit the deflections with our model galaxies. We'll
# fit it with two singular isothermal spheres at the same lens-plane, thus we should see how the absence of multi-plane
# ray tracing impacts the mass of the subhalo.


class DeflectionFitPhase(al.GalaxyFitPhase):
    def customize_priors(self, results):

        # You may wish to fix the first galaxy to its true centre / einstein radius

        #  self.galaxies.lens.mass.centre_0 = 0.0
        #  self.galaxies.lens.mass.centre_1 = 0.0
        #  self.galaxies.lens.mass.einstein_radius = 1.0

        # Adjusting the priors on the centre of galaxies away from (0.0", 0.0") is also a good idea.

        self.galaxies.subhalo.mass.centre_0 = af.GaussianPrior(mean=0.5, sigma=0.3)
        self.galaxies.subhalo.mass.centre_1 = af.GaussianPrior(mean=0.5, sigma=0.3)


phase = DeflectionFitPhase(
    phase_name="phase_galaxy_deflections_fit",
    galaxies=dict(
        lens=al.GalaxyModel(redshift=0.5, mass=al.mass_profiles.SphericalIsothermal),
        subhalo=al.GalaxyModel(redshift=0.5, mass=al.mass_profiles.SphericalIsothermal),
    ),
    use_deflections=True,
    sub_grid_size=4,
    mask_function=mask_function_circular,
    optimizer_class=af.MultiNest,
)


# Finally, when we run the phase, we now pass both deflection angle instrument's separately.
phase.run(galaxy_data=[data_y, data_x])

# If you check your output folder, you should see that this fit has been performed and visualization specific to a
# deflections fit is output.



import autofit as af
import autolens as al
from autolens.pipeline.phase import phase_imaging
from autolens.model.galaxy import galaxy_model as gm

import time

# In this tutorial, I want to show you 'MultiNest black magic'. Basically, there are ways to get
# MultiNest to run fast. Really fast. 30-40x faster than all of the previous tutorials!

# However, it risky, soo its important you develop an intuition for how this black magic works, so that you
# know when it is and isn't appropriate to use it.

# But, before we think about that, lets run two phase's, one without black magic and one with it. These runs
# will use the same prior config files (see 'chapter_2_lens_modeling/configs/7_multnest_black_magic'), thus any speed
# up in our phase's is not due to prior tuning.

# You need to change the path below to the chapter 2 directory.
chapter_path = "/path/to/user/autolens_workspace/howtolens/chapter_2_lens_modeling/"
chapter_path = "/home/jammy/PycharmProjects/PyAutoLens/workspace/howtolens/chapter_2_lens_modeling/"

af.conf.instance = af.conf.Config(
    config_path=chapter_path + "configs/t7_multinest_black_magic",
    output_path=chapter_path + "output",
)

# This function simulates the image we'll fit in this tutorial. Unlike previous tutorial images, it includes
# the light-profile of the lens galaxy.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=0.1)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(130, 130), pixel_scale=0.1, sub_grid_size=2
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.8,
            phi=45.0,
            intensity=0.2,
            effective_radius=0.8,
            sersic_index=3.0,
        ),
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 0.0), intensity=0.2, effective_radius=0.2
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.1,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Simulate the CCD data.
ccd_data = simulate()
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Lets first run the phase without black magic, which is performed as we're now used to.

# A word of warning, this phase takes >1 hour to run... so if you get bored, skip the run cell below  and continue to the
# phase with black magic.

phase_normal = phase_imaging.PhaseImaging(
    phase_name="phase_t7_no_black_magic",
    galaxies=dict(
        lens=gm.GalaxyModel(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic,
            mass=al.mass_profiles.EllipticalIsothermal,
        ),
        source=gm.GalaxyModel(redshift=1.0, light=al.light_profiles.EllipticalSersic),
    ),
    optimizer_class=af.MultiNest,
)

# We're going to use the time module to time how long each MultiNest run takes. However, if you resume the MultiNest
# run from a previous job, this time won't be accurate. Fortunately, if you look in the folder
# 'howtolens/chapter_2_lens_modeling/output/7_no_black_magic') you'll find a file 'run_time' which gives the overall
# run-time including any resumes.
start = time.time()

# Lets run the phase - the run-time will be output to the output/t7_multinest_black_magic/

print(
    "MultiNest has begun running - checkout the workspace/howtolens/chapter_2_lens_modeling/output/t7_multinest_black_magic"
    " folder for live output of the results, images and lens model."
    " This Jupyter notebook cell with progress once MultiNest has completed - this could take some time!"
)

phase_normal_results = phase_normal.run(data=ccd_data)

print("MultiNest has finished run - you may now continue the notebook.")

# Lets check we get a reasonably good model and fit to the data.
al.lens_fit_plotters.plot_fit_subplot(
    fit=phase_normal_results.most_likely_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

print("Time without black magic = {}".format(time.time() - start))

# Now lets run the phase with black magic on, which will hopefully run a lot faster than the previous phase.

phase_black_magic = phase_imaging.PhaseImaging(
    phase_name="phase_t7_with_black_magic",
    galaxies=dict(
        lens=gm.GalaxyModel(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic,
            mass=al.mass_profiles.EllipticalIsothermal,
        ),
        source=gm.GalaxyModel(redshift=1.0, light=al.light_profiles.EllipticalSersic),
    ),
    optimizer_class=af.MultiNest,
)

# And herein lies the black magic. The changes to n_live_points and sampling efficiency are part of it, but its
# the constant efficiency mode where the real magic lies. However, lets not worry about whats happening just let, I will
# explain all in a moment.
phase_black_magic.optimizer.n_live_points = 60
phase_black_magic.optimizer.sampling_efficiency = 0.5
phase_black_magic.optimizer.const_efficiency_mode = True

# Reset our timer.
start = time.time()

# Lets run the phase - the run-time will be output to the output/t7_multinest_black_magic/
print(
    "MultiNest has begun running - checkout the workspace/howtolens/chapter_2_lens_modeling/output/t7_multinest_black_magic"
    " folder for live output of the results, images and lens model."
    " This Jupyter notebook cell with progress once MultiNest has completed - this could take some time!"
)

phase_black_magic_results = phase_black_magic.run(data=ccd_data)

print("MultiNest has finished run - you may now continue the notebook.")

# Does our use of black magic impact the quality of our fit to the data?
al.lens_fit_plotters.plot_fit_subplot(
    fit=phase_black_magic_results.most_likely_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

print("Time with black magic = {}".format(time.time() - start))

# And there we have it, a speed up of our non-linear search of at least x30! Something about constant efficiency mode
# has lead to a huge speed up in our MultiNest search. So, what happened? How did we get such a large increase in
# run speed?

# To begin, we should think a bit more about how MultiNest works. MultiNest first 'maps out' parameter
# space over large scales using a set of live points (the number of which is defined by n_live_points). From these
# points it assesses what it thinks parameter space looks like and where it thinks the highest likelihood regions of
# parameter space are. MultiNest then 'guesses' (on average) more lens models from these higher likelihood regions of
# parameter space, with the hope that its live points will slowly converge around the maximum likelihood solution(s).

# How fast does MultiNest try to converge around these solutions? That is set by its sampling_efficiency. For example,
# a sampling efficiency of 0.3 means that MultiNest targets that 30% of its sample will result in 'accepted' live points
# (e.g. that they successfully sample a likelihood above at least one existing live point). For an efficiency of 0.8,
# it'd do this 80% of the time. Clearly, the higher our efficiency, the faster MultiNest samples parameter space.

# However, if MultiNest is not confident it has a sufficiently good map of parameter space that it can begin
# to converge around solutions at the sampling efficiency, it will lower the efficiency so as to more
# thoroughly map out parameter space. This is what is happening with the black magic switched off - the sampling
# efficiency doesn't retain the input value of 0.5 (50%) but instead drops dramtically to values of <5% by the end
# of the analysis. No longer it took so long to run, it took a HUGE amount of samples!

# The thing is, MultiNest doesn't really need to drop its acceptance rate. Its simply confused by the noisy and
# unsmooth parameter space we sample during lens modeling. A non linear sampler like MultiNest is expecting to
# see a perfectly smooth parameter space with no stochastic variation between two points close to one another in
# parameter space (e.g. a parametric space defined by a smooth analytic function f(x, y, z) = x^2 + y^3 - 4z^2). The
# parameter space we sample in lens modeling is not at all smooth and this upsets MultiNest to the point that its
# sampling often grinds to a halt.

# Enter, constant efficiency sampling mode! This mode *forces* MultiNest to retain the sampling efficiency acceptance
# rate regardless of whether or not it *thinks* it has a sufficiently good mapping out of parameter space. This gives
# us the huge speed up (as we saw for the black magic phase above), whilst ensuring we still compute an accurate lens
# model (because MultiNest had mapped out parameter space well enough, it just didn't know it). Therefore, MultiNest
# black magic is us 'tricking' MultiNest into not worrying too much about how thoroughly it samples parameter
# space and its an extremely powerful tool to keep run-times with PyAutoLens manageable.

# Of course, there are caveats and care must be taken. When we use constant efficiency mode, there is now the
# possibility that MultiNest will converge on a local maxima in parameter space and not be aware of it. We can see
# this by aggresively increasing the sampling efficiency and reducing the number of live points.

phase_too_much_black_magic = phase_imaging.PhaseImaging(
    phase_name="phase_t7_with_too_much_black_magic",
    galaxies=dict(
        lens=gm.GalaxyModel(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic,
            mass=al.mass_profiles.EllipticalIsothermal,
        ),
        source=gm.GalaxyModel(redshift=1.0, light=al.light_profiles.EllipticalSersic),
    ),
    optimizer_class=af.MultiNest,
)

phase_too_much_black_magic.optimizer.n_live_points = 20
phase_too_much_black_magic.optimizer.sampling_efficiency = 0.9
phase_too_much_black_magic.optimizer.const_efficiency_mode = True

# Reset our timer.
start = time.time()

# Lets run the phase - the run-time will be output to the output/t7_multinest_black_magic/
print(
    "MultiNest has begun running - checkout the workspace/howtolens/chapter_2_lens_modeling/output/t7_multinest_black_magic"
    " folder for live output of the results, images and lens model."
    " This Jupyter notebook cell with progress once MultiNest has completed - this could take some time!"
)

phase_too_much_black_magic_results = phase_too_much_black_magic.run(data=ccd_data)

print("MultiNest has finished run - you may now continue the notebook.")

al.lens_fit_plotters.plot_fit_subplot(
    fit=phase_too_much_black_magic_results.most_likely_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

print("Time with too much black magic = {}".format(time.time() - start))

# The phase ran super fast, but it gave us the incorrect lens model! We must use black magic with
# care!

# So, when should we use black magic and when shouldn't we? I generally follow the guidelines below:

# 1) When the dimensionality of parameter space is small < ~15 parameters.

# 2) If the parameter space is > ~15 parameters, when the priors on the majority of model parameters are initialized
#    using Gaussian priors centred on an accurate model.

# 3) When the lens model doesn't have high dimensionality degeneracies between different parameter (We'll expand on
#    this in later chapters).

# Finally, its worth emphasizing that when we cover pipelines in chapter 3 that black magic is extremely powerful.
# As we discussed in the previous tutorial, the whole premise of pipelines is we 'initialize' the lens model using a
# less accurate but more efficienct analysis, and worry about getting the results 'perfect' at the end. Thus, we'll see
# that in pipelines the early phases nearly always run in constant efficiency mode.



import autolens as al

grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=0.05, sub_grid_size=16
)

sie = al.mass_profiles.EllipticalIsothermalKormann(
    centre=(0.0, 0.0), einstein_radius=2.0, axis_ratio=0.2, phi=45.0
)
#  sie = al.mass_profiles.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0, axis_ratio=0.2, phi=45.0)

convergence_via_jacobian = sie.convergence_via_jacobian_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)
al.array_plotters.plot_array(
    array=convergence_via_jacobian,
    title="Kormann via Deflections & Jacobian",
    norm_max=1.2,
)

convergence = sie.convergence_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)
al.array_plotters.plot_array(
    array=convergence, title="Kormann Convergence", norm_max=1.2
)

deflections_via_jacobian = sie.deflections_via_potential_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)[:, :, 0]
deflections_via_jacobian = al.ScaledSquarePixelArray(
    array=deflections_via_jacobian, pixel_scale=0.05
)
al.array_plotters.plot_array(
    array=deflections_via_jacobian, title="Kormann via Deflections via Potential"
)

deflections = sie.deflections_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)[:, :, 0]
deflections = al.ScaledSquarePixelArray(array=deflections, pixel_scale=0.05)
al.array_plotters.plot_array(array=deflections, title="Kormann Deflections")



import autofit as af
import autolens as al

import numpy as np
import os

# This tool allows one to mask a bespoke mask for a given image of a strong lens, which can then be loaded
# before a pipeline is run and passed to that pipeline so as to become the default masked used by a phase (if a mask
# function is not passed to that phase).

# This tool creates an irmask, which can form any shape and is not restricted to circles, annuli, ellipses,
# etc. This mask is created as follows:

# 1) Blur the observed image with a Gaussian kernel of specified FWHM.
# 2) Compute the absolute S/N map of that blurred image and the noise-map.
# 3) Create the mask for all pixels with a S/N above a theshold value.

# The following parameters determine the behaviour of this function:
blurring_gaussian_sigma = 0.02  # The FWHM of the Gaussian the image is blurred with
signal_to_noise_threshold = (
    10.0
)  # The threshold S/N value the blurred image must be above to not be masked.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# Setup the path to the config folder, using the workspace path.
config_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
af.conf.instance = af.conf.Config(
    config_path=config_path + "config", output_path=workspace_path + "output"
)

# The 'data name' is the name of the data folder and 'data_name' the folder the mask is stored in, e.g,
# the mask will be output as '/workspace/data/data_type/data_name/mask.fits'.
data_type = "example"
data_name = "lens_sie__source_sersic"

# Create the path where the mask will be output, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

# If you use this tool for your own instrument, you *must* double check this pixel scale is correct!
pixel_scale = 0.1

# First, load the CCD imaging data and noise-map, so that the mask can be plotted over the strong lens image.
image = al.load_image(
    image_path=data_path + "image.fits", image_hdu=0, pixel_scale=pixel_scale
)

noise_map = al.load_image(
    image_path=data_path + "noise_map.fits", image_hdu=0, pixel_scale=pixel_scale
)

# Create the 2D Gaussian that the image is blurred with. This blurring is an attempt to smooth over noise in the
# image, which will lead unmasked values with in individual pixels if not smoothed over correctly.
blurring_gaussian = al.PSF.from_gaussian(
    shape=(31, 31), pixel_scale=pixel_scale, sigma=blurring_gaussian_sigma
)

blurred_image = blurring_gaussian.convolve(image)

blurred_image = al.ScaledSquarePixelArray(array=blurred_image, pixel_scale=pixel_scale)

al.array_plotters.plot_array(array=blurred_image)

# Now compute the absolute signal-to-noise map of this blurred image, given the noise-map of the observed instrument.
blurred_signal_to_noise_map = blurred_image / noise_map

blurred_signal_to_noise_map = al.ScaledSquarePixelArray(
    array=blurred_signal_to_noise_map, pixel_scale=pixel_scale
)

al.array_plotters.plot_array(array=blurred_signal_to_noise_map)

# Now create the mask in sall pixels where the signal to noise is above some threshold value.
mask = np.where(blurred_signal_to_noise_map > signal_to_noise_threshold, False, True)

mask = al.Mask(array=mask, pixel_scale=pixel_scale)

al.array_plotters.plot_array(array=image, mask=mask)

# Now we're happy with the mask, lets output it to the data folder of the lens, so that we can load it from a .fits
# file in our pipelines!
al.output_mask_to_fits(mask=mask, mask_path=data_path + "mask.fits", overwrite=True)



from autolens.array import grids
from autolens.plotters import grid_plotters

# In this example, we'll create grids of Cartesian (y,x) coordinates, representing the arc-second coordinate grid of
# an observed data-set (e.g. ccd imaging).

# In PyAutoLens, a grid is a set of two-dimensional (y,x) coordinates (in arc-seconds) that are deflected
# and traced by a strong lensing system.

# This coordinate grid is aligned with the image we analyze, such that each coordinate on a  grid maps to the
# centre of each image-pixel. Lets make a grid using 100 x 100 pixels, with a pixel scale (arcsecond-to-pixel
# conversion factor) of 0.05", giving us a a 5" x 5" grid.
grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=0.05
)

# First, lets plot this grid, which shows that it is a fairly bland uniform grid of dots.
grid_plotters.plot_grid(grid=grid, title="Fairly Bland Uniform Grid Of Dots")

# We can print each coordinate of this grid, revealing that it consists of a set of arc-second coordinates
# (where the spacing between each coordinate corresponds to the 'pixel_scale' of 0.05" defined above)
print("(y,x) pixel 0:")
print(grid[0])
print("(y,x) pixel 1:")
print(grid[1])
print("(y,x) pixel 2:")
print(grid[2])
print("(y,x) pixel 100:")
print(grid[100])
print("etc.")

# (PyAutoLens flattens the NumPy array to 1D to save memory, however it consists of all
# 10000 pixel coordinates on our 2D grid)
print(grid.shape)

# A grid can be created with a sub-grid, which splits each pixel on the grid into sub-pixel of size (sub_grid_size
# x sub_grid_size) and which are used to perform calculations more accurately:
grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
)

# We specified a sub_grid_size of 2 above, therefore we expect 4 (2 x 2) times more sub-pixels than pixels
# (which like the grid are again flattened to 1D).
print(grid.shape)

# The first four pixels of our sub-grid correspond to the first four sub-pixels, which form a sub-grid inside the
# first grid pixel, as follows:

# pixel 1:
#              _________
#              |         |
#              |         |  o = (y,x) centre of
# y = 2.475"   |    o    |       grid coordinate.
#              |         |
#              |_________|
#              x = -2.475

# Sub-pixels 1, 2, 3 & 4:
#               _________
#              |         |
#              |  o   o  |  o = (y,x) centre of sub-
# y = 2.475"   |         |       grid coordinates.
#              |  o   o  |
#              |_________|
#              x = -2.475

# The sub-pixels coordinate are spaced uniformly between the pixel's edges
# (which are at y = (2.45", 2.5") and x = (-2.5". -2.45") )
print("(y,x) sub-pixel 0 (of pixel 0):")
print(grid[0])
print("(y,x) sub-pixel 1 (of pixel 0):")
print(grid[1])
print("(y,x) sub-pixel 2 (of pixel 0):")
print(grid[2])
print("(y,x) sub-pixel 3 (of pixel 0):")
print(grid[3])

# The sub-grid then continues on to the next grid pixels (and so on)
print("(y,x) sub-pixel 0 (of pixel 1):")
print(grid[4])
print("(y,x) sub-pixel 0 (of pixel 1):")
print(grid[5])
print("etc.")

# Congratulations, you've completed your first PyAutoLens tutorial! Before moving on to the next one, experiment with
# PyAutoLens by doing the following:

# 1) Change the pixel-scale of the grids - what happens to the Cartesian coordinates?
# 2) Change the resolution of the grids - what happens Cartesian coordinates?
# 3) Change the sub-grid size - what do you note about the pixel-scale of sub pixels?



import autofit as af
import autolens as al

############## THIS PIPELINE IS A WORK IN PROGRESS AND I DO NOT RECOMMEND YOU TRY TO USE IT TO DO LENS MODELING ########

# In this pipeline, we'll perform an advanced analysis which fits a lens galaxy with three surrounding line-of-sight \
# galaxies (which are at different redshifts and thus define a multi-plane stronog lens configuration). The source \
# galaxy will be modeled using a parametric light profile in the initial phases, but switch to an inversion in the \
# later phases.

# For efficiency, we will subtract the lens galaxy's light and line-of-sight galaxies light, and then only fit a
# foreground subtracted image. This means we can use an annular mask tailored to the source galaxy's light, which we
# have setup already using the 'tools/mask_maker.py' scripts.

# This leads to a 5 phase pipeline:

# Phase 1) Fit and subtract the light profile of the lens galaxy (elliptical Sersic) and each line-of-sight
#          galaxy (spherical Sersic).

# Phase 2) Use this lens subtracted image to fit the lens galaxy's mass (SIE) and source galaxy's light (Sersic),
#          thereby omitting multi-plane ray-tracing.

# Phase 3) Initialize the resolution and regularization coefficient of the inversion using the best-fit lens model from
#          phase 2.

# Phase 4) Refit the lens galaxy's light and mass models using an inversion (with priors from the above phases). The
#          mass profile of the 3 line-of-sight galaxies is also included (as SIS profiles), but a single lens plane
#          is assumed.

# Phase 5) Fit the lens galaxy, line-of-sight galaxies and source-galaxy using multi-plane ray-tracing, where the
#          redshift of each line-of-sight galaxy is included in the non-linear search as a free parameter. This phase
#          uses priors initialized from phase 4.

# Phase 5 risks the inversion falling into the systematic solutions where the mass of the lens model is too high,
# resulting in the inversion reconstructing the image as a demagnified version of itself. This is because the \
# line-of-sight halos each have a mass, which during the modeling can go to very high values.

# For this reason, we include positions (drawn using the 'tools/positions.py' script) to prevent these solutions from
# existing in parameter space.


def make_pipeline(phase_folders=None):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_multi_plane"
    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings()

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    # In phase 1, we will:

    # 1) Subtract the light of the main lens galaxy (located at (0.0", 0.0")) and the light of each line-of-sight
    # galaxy (located at (4.0", 4.0"), (3.6", -5.3") and (-3.1", -2.4"))

    class LensPhase(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.lens.light.centre_0 = af.GaussianPrior(mean=0.0, sigma=0.1)
            self.galaxies.lens.light.centre_1 = af.GaussianPrior(mean=0.0, sigma=0.1)
            self.galaxies.los0.light.centre_0 = af.GaussianPrior(mean=4.0, sigma=0.1)
            self.galaxies.los0.light.centre_1 = af.GaussianPrior(mean=4.0, sigma=0.1)
            self.galaxies.los1.light.centre_0 = af.GaussianPrior(mean=3.6, sigma=0.1)
            self.galaxies.los1.light.centre_1 = af.GaussianPrior(mean=-5.3, sigma=0.1)
            self.galaxies.los2.light.centre_0 = af.GaussianPrior(mean=-3.1, sigma=0.1)
            self.galaxies.los2.light.centre_1 = af.GaussianPrior(mean=-2.4, sigma=0.1)

    phase1 = LensPhase(
        phase_name="phase_1__light_subtraction",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(light=al.light_profiles.EllipticalSersic),
            los_0=al.GalaxyModel(light=al.light_profiles.SphericalSersic),
            los_1=al.GalaxyModel(light=al.light_profiles.SphericalSersic),
            los_2=al.GalaxyModel(light=al.light_profiles.SphericalSersic),
        ),
        optimizer_class=af.MultiNest,
    )

    # Customize MultiNest so it runs fast
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2
    phase1.optimizer.const_efficiency_mode = True

    # In phase 2, we will:

    # 1) Fit this foreground subtracted image, using an SIE+Shear mass model and Sersic source.
    # 2) Use the input positions to resample inaccurate mass models.

    class LensSubtractedPhase(al.PhaseImaging):
        def modify_image(self, image, results):
            return (
                image
                - results.from_phase(
                    "phase_1__light_subtraction"
                ).unmasked_lens_power_lawane_model_image
            )

        def customize_priors(self, results):

            self.galaxies.lens.mass.centre_0 = af.GaussianPrior(mean=0.0, sigma=0.1)
            self.galaxies.lens.mass.centre_1 = af.GaussianPrior(mean=0.0, sigma=0.1)

    phase2 = LensSubtractedPhase(
        phase_name="phase_2__source_parametric",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(mass=al.mass_profiles.EllipticalIsothermal),
            source=al.GalaxyModel(light=al.light_profiles.EllipticalSersic),
        ),
        positions_threshold=0.3,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.2
    phase2.optimizer.const_efficiency_mode = True

    # In phase 3, we will:

    #  1) Fit the foreground subtracted image using a source-inversion instead of parametric source, using lens galaxy
    #     priors from phase 2.

    class InversionPhase(al.PhaseImaging):
        def modify_image(self, image, results):
            return (
                image
                - results.from_phase(
                    "phase_1__light_subtraction"
                ).unmasked_lens_power_lawane_model_image
            )

        def customize_priors(self, results):

            self.galaxies.lens = results[1].constant.lens

            self.galaxies.source.pixelization.shape_0 = af.UniformPrior(
                lower_limit=20.0, upper_limit=45.0
            )
            self.galaxies.source.pixelization.shape_1 = af.UniformPrior(
                lower_limit=20.0, upper_limit=45.0
            )

    phase3 = InversionPhase(
        phase_name="phase_3__inversion_init",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(mass=al.mass_profiles.EllipticalIsothermal),
            source=al.GalaxyModel(
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    # Customize MultiNest so it runs fast
    phase3.optimizer.n_live_points = 10
    phase3.optimizer.sampling_efficiency = 0.5
    phase3.optimizer.const_efficiency_mode = True

    # In phase 4, we will:

    #  1) Fit the foreground subtracted image using a source-inversion with parameters initialized from phase 3.
    #  2) Initialize the lens galaxy priors using the results of phase 2, and include each line-of-sight galaxy's mass
    #     contribution (keeping its centre fixed to the light profile centred inferred in phase 1). This will assume
    #     a single lens plane for all line-of-sight galaxies.
    #  3) Use the input positions to resample inaccurate mass models.

    class SingleLensPlanePhase(al.PhaseImaging):
        def modify_image(self, image, results):
            return (
                image
                - results.from_phase(
                    "phase_1__light_subtraction"
                ).unmasked_lens_power_lawane_model_image
            )

        def customize_priors(self, results):

            self.galaxies.lens.mass = results.from_phase(
                "phase_3__inversion_init"
            ).variable.lens.mass

            self.galaxies.los_0.mass.centre_0 = results[0].constant.los_0.light.centre_0
            self.galaxies.los_0.mass.centre_1 = results[0].constant.los_0.light.centre_1
            self.galaxies.los_1.mass.centre_0 = results[0].constant.los_1.light.centre_0
            self.galaxies.los_1.mass.centre_1 = results[0].constant.los_1.light.centre_1
            self.galaxies.los_2.mass.centre_0 = results[0].constant.los_2.light.centre_0
            self.galaxies.los_2.mass.centre_1 = results[0].constant.los_2.light.centre_1

            self.galaxies.source = results.from_phase(
                "phase_3__inversion_init"
            ).variable.source

    phase4 = SingleLensPlanePhase(
        phase_name="phase_4__single_plane",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(mass=al.mass_profiles.EllipticalIsothermal),
            los_0=al.GalaxyModel(mass=al.mass_profiles.SphericalIsothermal),
            los_1=al.GalaxyModel(mass=al.mass_profiles.SphericalIsothermal),
            los_2=al.GalaxyModel(mass=al.mass_profiles.SphericalIsothermal),
            source=al.GalaxyModel(
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        positions_threshold=0.3,
        optimizer_class=af.MultiNest,
    )

    # Customize MultiNest so it runs fast
    phase4.optimizer.n_live_points = 60
    phase4.optimizer.sampling_efficiency = 0.2
    phase4.optimizer.const_efficiency_mode = True

    # In phase 5, we will fit the foreground subtracted image using a multi-plane ray tracer. This means that the
    # redshift of each line-of-sight galaxy is included as a free parameter (we assume the lens and source redshifts
    # are known).

    class MultiPlanePhase(al.PhaseImaging):
        def modify_image(self, image, results):
            return (
                image
                - results.from_phase(
                    "phase_1__light_subtraction"
                ).unmasked_lens_power_lawane_model_image
            )

        def customize_priors(self, results):

            self.galaxies.lens = results.from_phase(
                "phase_4__single_plane"
            ).variable.lens

            self.galaxies.los_0.mass = results.from_phase(
                "phase_4__single_plane"
            ).variable.los_0.mass
            self.galaxies.los_1.mass = results.from_phase(
                "phase_4__single_plane"
            ).variable.los_1.mass
            self.galaxies.los_2.mass = results.from_phase(
                "phase_4__single_plane"
            ).variable.los_2.mass

            self.galaxies.source = results.from_phase(
                "phase_4__single_plane"
            ).variable.source

    phase5 = MultiPlanePhase(
        phase_name="phase_5_multi_plane",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(mass=al.mass_profiles.EllipticalIsothermal),
            los_0=al.GalaxyModel(
                mass=al.mass_profiles.SphericalIsothermal, variable_redshift=True
            ),
            los_1=al.GalaxyModel(
                mass=al.mass_profiles.SphericalIsothermal, variable_redshift=True
            ),
            los_2=al.GalaxyModel(
                mass=al.mass_profiles.SphericalIsothermal, variable_redshift=True
            ),
            source=al.GalaxyModel(
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        positions_threshold=0.3,
        optimizer_class=af.MultiNest,
    )

    # Customize MultiNest so it runs fast
    phase5.optimizer.n_live_points = 60
    phase5.optimizer.sampling_efficiency = 0.2
    phase5.optimizer.const_efficiency_mode = True

    return al.PipelineImaging(pipeline_name, phase1, phase2, phase3, phase4, phase5)



# Up to now, all the images we've fitted had one lens galaxy. However, we saw in chapter 1 that our lens plane can consist
# of multiple galaxies which each contribute to the strong lensing. Multi-galaxy systems are challenging to
# model, because they add an extra 5-10 parameters to the non-linear search and, more problematically, the
# degeneracies between the mass-profiles of the two galaxies can be severe.

# However, we can still break their analysis down using a pipeline and give ourselves a shot at getting a good
# lens model. Here, we're going to fit a double lens system, fitting as much about each individual lens galaxy before
# fitting them simultaneously.

# Up to now, I've put a focus on pipelines being general. The pipeline we write in this example is going to be
# the opposite - specific to the image we're modeling. Fitting multiple lens galaxies is really difficult and
# writing a pipeline that we can generalize to many lenses isn't currently possible with PyAutoLens.

### AUTOFIT + CONFIG SETUP ###

import autofit as af

# Setup the path to the workspace, using by filling in your path below.
workspace_path = "/path/to/user/autolens_workspace/"
workspace_path = "/home/jammy/PycharmProjects/PyAutoLens/workspace/"

# Setup the path to the config folder, using the workspace path.
config_path = workspace_path + "config"

# Use this path to explicitly set the config path and output path.
af.conf.instance = af.conf.Config(
    config_path=config_path, output_path=workspace_path + "output"
)

### AUTOLENS + DATA SETUP ###

import autolens as al

# This rather long simulate function generates an image with two strong lens galaxies.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05
    )

    lens_galaxy_0 = al.Galaxy(
        redshift=0.5,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, -1.0),
            axis_ratio=0.8,
            phi=55.0,
            intensity=0.1,
            effective_radius=0.8,
            sersic_index=2.5,
        ),
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(1.0, 0.0), axis_ratio=0.7, phi=45.0, einstein_radius=1.0
        ),
    )

    lens_galaxy_1 = al.Galaxy(
        redshift=0.5,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 1.0),
            axis_ratio=0.8,
            phi=100.0,
            intensity=0.1,
            effective_radius=0.6,
            sersic_index=3.0,
        ),
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(-1.0, 0.0), axis_ratio=0.8, phi=90.0, einstein_radius=0.8
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.SphericalExponential(
            centre=(0.05, 0.15), intensity=0.2, effective_radius=0.5
        ),
    )

    tracer = al.Tracer.from_galaxies(
        galaxies=[lens_galaxy_0, lens_galaxy_1, source_galaxy]
    )

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Lets Simulate the CCD data we'll fit, which is a new image, finally!
ccd_data = simulate()
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Looking at the image, there are clearly two blobs of light corresponding to our two lens galaxies. The source's light
# is also pretty complex - the arcs don't posses the rotational symmetry we're used to seeing up to now. Multi-galaxy
# ray-tracing is just a lot more complicated, which means so is modeling it!

# So, how can we break the lens modeling up? As follows:

# 1) Fit and subtract the light of each lens galaxy individually - this will require some careful masking but is doable.
# 2) Use these results to initialize each lens galaxy's mass-profile.

# So, with this in mind, we've written a pipeline composed of 4 phases:

# Phase 1) Fit the light profile of the lens galaxy on the left of the image, at coordinates (0.0", -1.0").
# Phase 2) Fit the light profile of the lens galaxy on the right of the image, at coordinates (0.0", 1.0").
# Phase 3) Use this lens-subtracted image to fit the source galaxy's light. The mass-profiles of the two lens galaxies
#          can use the results of phases 1 and 2 to initialize their priors.
# Phase 4) Fit all relevant parameters simultaneously, using priors from phases 1, 2 and 3.

# Again, before we checkout the pipeline, lets import it, and get it running.
from workspace.howtolens.chapter_3_pipelines import tutorial_2_pipeline_x2_lens_galaxies

pipeline_x2_galaxies = tutorial_2_pipeline_x2_lens_galaxies.make_pipeline(
    phase_folders=["howtolens", "c3_t2_x2_galaxies"]
)

pipeline_x2_galaxies.run(data=ccd_data)

# Now, read through the '_tutorial_2_pipeline_x2_galaxies.py_' pipeline, to get a complete view of how it works.
# Once you've done that, come back here and we'll wrap up this tutorial.


# And, we're done. This pipeline takes a while to run, as is the nature of multi-galaxy modeling. Nevertheless, the
# techniques we've learnt above can be applied to systems with even more galaxies, albeit the increases in parameters
# will slow down the non-linear search. Here are some more Q&A's

# 1) This system had two very similar lens galaxies, with comparable amounts of light and mass. How common is this?
#    Does it make it harder to model them?

#    Typically, a 2 galaxy system has 1 massive galaxy (that makes up some 80%-90% of the overall light and mass),
#    accompanied by a smaller satellite. The satellite can't be ignored - it impacts the ray-tracing in a measureable
#    way, but its a lot less degenerate with the 'main' lens galaxy. This means we can often model the  satellite
#    with much simpler profiles (e.g. spherical profiles). So yes, multi-galaxy systems can often be easier to model.

# 2) It got pretty confusing passing all those priors towards the end of the pipeline there, didn't it?

#    It does get confusing, I won't lie. This is why we made galaxies named objects - so that we could call them the
#    'left_lens' and 'right_lens'. It still requires caution when writing the pipeline, but goes to show that if
#    you name your galaxies sensibly you should be able to avoid errors, or spot them quickly when you make them.



from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.array import grids
from autolens.lens import ray_tracing
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.lens.plotters import ray_tracing_plotters
from autolens.data.plotters import ccd_plotters

# In this example, we'll use the 'ccd' module to 'simulate' ccd imaging of a strong lens made
# using a tracer. By simulate, we mean that it will appear as if we had observed it using a real telescope,
# with this example making an image representative of Hubble Space Telescope imaging.

# To simulate an image, we need to model the telescope's optics. We'll do this by convolving the image with a
# Point-Spread Function, which we can simulate as a Gaussian using the abstract-data module.
psf = abstract_data.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=0.1)

# To simulate ccd instrument, we use a grid, like usual.
grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=0.1, sub_grid_size=2
)

# Now, lets setup our lens galaxy, source galaxy and tracer.
lens_galaxy = g.Galaxy(
    redshift=0.5,
    mass=mp.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
)

source_galaxy = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)

tracer = ray_tracing.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

# Lets look at the tracer's image - this is the image we'll be simulating.
ray_tracing_plotters.plot_profile_image(tracer=tracer, grid=grid)

# To Simulate the CCD data, we don't use the image plotted above. Instead, we use an image which has been generated
# specifically for simulating an image, which pads the array it is computed on based on the shape of the PSF we
# convolve the image with. This ensures edge-effects do not degrade our simulation's PSF convolution.
normal_image = tracer.profile_image_from_grid(grid=grid, return_in_2d=True)
padded_image = tracer.padded_profile_image_2d_from_grid_and_psf_shape(
    grid=grid, psf_shape=psf.shape
)
print(normal_image.shape)
print(padded_image.shape)

# Now, to simulate the ccd imaging data, we pass the tracer and grid to the ccd module's simulate
# function. This adds the following effects to the image:

# 1) Telescope optics: Using the Point Spread Function above.
# 2) The Background Sky: Although the image that is returned is automatically background sky subtracted.
# 3) Poisson noise: Due to the background sky, lens galaxy and source galaxy Poisson photon counts.

ccd_data = ccd.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=0.1,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)

# Lets plot the image - we can see the image has been blurred due to the telescope optics and noise has been added.
ccd_plotters.plot_image(ccd_data=ccd_data)

# Finally, lets output these files to.fits files, we'll begin to analyze them in the next tutorial!
chapter_path = "/path/to/AutoLens/workspace/howtolens/chapter_1_introduction"
chapter_path = (
    "/home/jammy/PycharmProjects/PyAutoLens/workspace/howtolens/chapter_1_introduction/"
)

# The data path specifies where the data is output, this time in the directory 'chapter_path/data'
data_path = chapter_path + "data/"

# Now output our simulated data to hard-disk.
ccd.output_ccd_data_to_fits(
    ccd_data=ccd_data,
    image_path=data_path + "image.fits",
    noise_map_path=data_path + "noise_map.fits",
    psf_path=data_path + "psf.fits",
    overwrite=True,
)



from autofit import conf
import os

workspace_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=workspace_path + "output"
)

from autolens.model.profiles import mass_profiles as mp

cosmology_util.angular_diameter_distance_between_redshifts_from_redshifts_and_cosmlology()


nfw = al.mass_profiles.SphericalTruncatedNFWChallenge(kappa_s=0.033, scale_radius=0.169)
# nfw = al.mass_profiles.SphericalTruncatedNFWChallenge(kappa_s=0.017, scale_radius=0.44)

sigma_cr = 1940654909.4133248
cos_den = 262.30319684750657

mass = nfw.mass_at_200_for_units(
    critical_surface_density=sigma_cr, cosmic_average_density=cos_den
)
con = nfw.concentration_for_units(
    critical_surface_density=sigma_cr, cosmic_average_density=cos_den
)
print("level 0 - large_hi_sn_system_1")
print("mass = {:.4e}, concentration = {:.4f} \n".format(mass, con))

nfw = al.mass_profiles.SphericalTruncatedNFWChallenge(kappa_s=0.011, scale_radius=0.819)
nfw = al.mass_profiles.SphericalTruncatedNFWChallenge(kappa_s=0.016, scale_radius=0.389)

sigma_cr = 1940654909.4133248
cos_den = 262.30319684750657

mass = nfw.mass_at_200_for_units(
    critical_surface_density=sigma_cr, cosmic_average_density=cos_den
)
con = nfw.concentration_for_units(
    critical_surface_density=sigma_cr, cosmic_average_density=cos_den
)
print("level 0 - small_hi_sn_system_1")
print("mass = {:.4e}, concentration = {:.4f} \n".format(mass, con))



from autolens import decorator_util
import numpy as np


@decorator_util.jit()
def data_vector_from_blurred_mapping_matrix_and_data(
    blurred_mapping_matrix, image_1d, noise_map_1d
):
    """Compute the hyper_galaxies vector *D* from a blurred mapping_util matrix *f* and the 1D image *d* and 1D noise-map *\sigma* \
    (see Warren & Dye 2003).
    
    Parameters
    -----------
    blurred_mapping_matrix : ndarray
        The matrix representing the blurred mappings between sub-grid pixels and pixelization pixels.
    image_1d : ndarray
        Flattened 1D array of the observed image the inversion is fitting.
    noise_map_1d : ndarray
        Flattened 1D array of the noise-map used by the inversion during the fit.
    """

    mapping_shape = blurred_mapping_matrix.shape

    data_vector = np.zeros(mapping_shape[1])

    for mask_1d_index in range(mapping_shape[0]):
        for pixelization_1d_index in range(mapping_shape[1]):
            data_vector[pixelization_1d_index] += (
                image_1d[mask_1d_index]
                * blurred_mapping_matrix[mask_1d_index, pixelization_1d_index]
                / (noise_map_1d[mask_1d_index] ** 2.0)
            )

    return data_vector


def curvature_matrix_from_blurred_mapping_matrix(blurred_mapping_matrix, noise_map_1d):
    """Compute the curvature matrix *F* from a blurred mapping_util matrix *f* and the 1D noise-map *\sigma* \
     (see Warren & Dye 2003).

    Parameters
    -----------
    blurred_mapping_matrix : ndarray
        The matrix representing the blurred mappings between sub-grid pixels and pixelization pixels.
    noise_map_1d : ndarray
        Flattened 1D array of the noise-map used by the inversion during the fit.
    """

    flist = np.zeros(blurred_mapping_matrix.shape[1])
    iflist = np.zeros(blurred_mapping_matrix.shape[1], dtype="int")
    return curvature_matrix_from_blurred_mapping_matrix_jit(
        blurred_mapping_matrix, noise_map_1d, flist, iflist
    )


@decorator_util.jit()
def curvature_matrix_from_blurred_mapping_matrix_jit(
    blurred_mapping_matrix, noise_map_1d, flist, iflist
):
    """Compute the curvature matrix *F* from a blurred mapping_util matrix *f* and the 1D noise-map *\sigma* \
    (see Warren & Dye 2003).

    Parameters
    -----------
    blurred_mapping_matrix : ndarray
        The matrix representing the blurred mappings between sub-grid pixels and pixelization pixels.
    noise_map_1d : ndarray
        Flattened 1D array of the noise-map used by the inversion during the fit.
    flist : ndarray
        NumPy array of floats used to store mappings for efficienctly calculation.
    iflist : ndarray
        NumPy array of integers used to store mappings for efficienctly calculation.
    """
    curvature_matrix = np.zeros(
        (blurred_mapping_matrix.shape[1], blurred_mapping_matrix.shape[1])
    )

    for mask_1d_index in range(blurred_mapping_matrix.shape[0]):
        index = 0
        for pixelization_1d_index in range(blurred_mapping_matrix.shape[1]):
            if blurred_mapping_matrix[mask_1d_index, pixelization_1d_index] > 0.0:
                flist[index] = (
                    blurred_mapping_matrix[mask_1d_index, pixelization_1d_index]
                    / noise_map_1d[mask_1d_index]
                )
                iflist[index] = pixelization_1d_index
                index += 1

        if index > 0:
            for i1 in range(index):
                for j1 in range(index):
                    ix = iflist[i1]
                    iy = iflist[j1]
                    curvature_matrix[ix, iy] += flist[i1] * flist[j1]

    for i in range(blurred_mapping_matrix.shape[1]):
        for j in range(blurred_mapping_matrix.shape[1]):
            curvature_matrix[i, j] = curvature_matrix[j, i]

    return curvature_matrix


@decorator_util.jit()
def reconstructed_data_vector_from_blurred_mapping_matrix_and_solution_vector(
    blurred_mapping_matrix, solution_vector
):
    """ Compute the reconstructed hyper_galaxies vector from the blurrred mapping_util matrix *f* and solution vector *S*.

    Parameters
    -----------
    blurred_mapping_matrix : ndarray
        The matrix representing the blurred mappings between sub-grid pixels and pixelization pixels.

    """
    reconstructed_data_vector = np.zeros(blurred_mapping_matrix.shape[0])
    for i in range(blurred_mapping_matrix.shape[0]):
        for j in range(solution_vector.shape[0]):
            reconstructed_data_vector[i] += (
                solution_vector[j] * blurred_mapping_matrix[i, j]
            )

    return reconstructed_data_vector


def pixelization_residual_map_from_pixelization_values_and_reconstructed_data_1d(
    pixelization_values,
    reconstructed_data_1d,
    sub_mask_1d_index_to_mask_1d_index,
    pixelization_1d_index_to_all_sub_mask_1d_indexes,
):

    pixelization_residuals = np.zeros(
        shape=len(pixelization_1d_index_to_all_sub_mask_1d_indexes)
    )

    reconstructed_data_1d = reconstructed_data_1d

    for pixelization_1d_index, sub_mask_1d_indexes in enumerate(
        pixelization_1d_index_to_all_sub_mask_1d_indexes
    ):
        for sub_mask_1d_index in sub_mask_1d_indexes:
            mask_1d_index = sub_mask_1d_index_to_mask_1d_index[sub_mask_1d_index]
            residual = (
                reconstructed_data_1d[mask_1d_index]
                - pixelization_values[pixelization_1d_index]
            )
            pixelization_residuals[pixelization_1d_index] += np.abs(residual)

    return pixelization_residuals


def pixelization_normalized_residual_map_from_pixelization_values_and_reconstructed_data_1d(
    pixelization_values,
    reconstructed_data_1d,
    noise_map_1d,
    sub_mask_1d_index_to_mask_1d_index,
    pixelization_1d_index_to_all_sub_mask_1d_indexes,
):

    pixelization_normalized_residuals = np.zeros(
        shape=len(pixelization_1d_index_to_all_sub_mask_1d_indexes)
    )

    reconstructed_data_1d = reconstructed_data_1d

    for pixelization_1d_index, sub_mask_1d_indexes in enumerate(
        pixelization_1d_index_to_all_sub_mask_1d_indexes
    ):
        for sub_mask_1d_index in sub_mask_1d_indexes:
            mask_1d_index = sub_mask_1d_index_to_mask_1d_index[sub_mask_1d_index]
            residual = (
                reconstructed_data_1d[mask_1d_index]
                - pixelization_values[pixelization_1d_index]
            )
            pixelization_normalized_residuals[pixelization_1d_index] += np.abs(
                (residual / noise_map_1d[mask_1d_index])
            )

    return pixelization_normalized_residuals


def pixelization_chi_squared_map_from_pixelization_values_and_reconstructed_data_1d(
    pixelization_values,
    reconstructed_data_1d,
    noise_map_1d,
    sub_mask_1d_index_to_mask_1d_index,
    pixelization_1d_index_to_all_sub_mask_1d_indexes,
):

    pixelization_chi_squareds = np.zeros(
        shape=len(pixelization_1d_index_to_all_sub_mask_1d_indexes)
    )

    reconstructed_data_1d = reconstructed_data_1d

    for pixelization_1d_index, sub_mask_1d_indexes in enumerate(
        pixelization_1d_index_to_all_sub_mask_1d_indexes
    ):
        for sub_mask_1d_index in sub_mask_1d_indexes:
            mask_1d_index = sub_mask_1d_index_to_mask_1d_index[sub_mask_1d_index]
            residual = (
                reconstructed_data_1d[mask_1d_index]
                - pixelization_values[pixelization_1d_index]
            )
            pixelization_chi_squareds[pixelization_1d_index] += (
                residual / noise_map_1d[mask_1d_index]
            ) ** 2.0

    return pixelization_chi_squareds
